/*******************************************************************************
 * Copyright (c) 2016 Dec. Verimag. All rights reserved.
 * @author Hang YU
*******************************************************************************/

#ifndef _GLPKINTER
#define _GLPKINTER
#include "glpkInterface.h"
#endif

GlpkInterface::GlpkInterface () 
  : _epsilon(0.1) {
  _glp = glp_create_prob() ;
}

GlpkInterface::~GlpkInterface () { 
  glp_delete_prob(_glp) ;
}

/*******************************************************************************
 * Loads a file in CPLEX LP format, and create a Polyhedron object
 * Note that the index of row and column of GLPK starts from 1
 * @para filepath the path of text file in CPLEX LP format
 * @return a Polyhedron object which contains all information of the polyhdedron 
*******************************************************************************/
Polyhedron GlpkInterface::LoadLP (char* filepath) {
  GlpkInterface glpkinter ;
  glp_read_lp(glpkinter._glp, NULL, filepath) ; // TODO error info
  int consNum = glp_get_num_rows(glpkinter._glp) ;
  int variNum = glp_get_num_cols(glpkinter._glp) ;
  Polyhedron poly = Polyhedron(consNum, variNum) ;
  int* tempIdx = new int[consNum+1] ;
  double* tempVal = new double[variNum+1] ;
  for (int i = 0; i < consNum; ++ i) {
    // tempVal should be initialized as 0, otherwise glpk bug
    std::fill_n(tempIdx, consNum+1, 0) ;
    std::fill_n(tempVal, variNum+1, 0) ;
    poly.SetConstant( i, glp_get_row_ub(glpkinter._glp, i+1) ) ;
    // the row index is i+1 because in GLPK the index starts from 1
    glp_get_mat_row(glpkinter._glp, i+1, tempIdx, tempVal) ;
    for (int j = 0, k = 1; j < variNum; ++ j) {
      // the col index is j+1 because in GLPK the index starts from 1
      if (j+1 == tempIdx[k]) {
        poly.SetCoef(i, j, tempVal[k]) ;
        ++ k ;
      }      
      else {
        poly.SetCoef(i, j, 0.0) ;
      }
    }
  }
  delete[] tempIdx ;
  delete[] tempVal ;

  return poly ;
}

/*******************************************************************************
 * Call simplex method in GLPK and compute the central point of the polyhedron
 * @para poly the given polyhedron
 * @return the central point  
*******************************************************************************/
Point GlpkInterface::GetCentralPoint (const Polyhedron& poly) {
  GlpkInterface glpkinter ;
  glp_set_obj_dir(glpkinter._glp, GLP_MIN) ;
  int consNum = poly.get_constraints_num() ;
  int variNum = poly.get_variable_num() ;
  glp_add_rows(glpkinter._glp, 2*consNum) ; 
  // the constraints: ax <= b
  // ||ax - b|| / ||a|| >= r1  =>  ax + ||a||r1 <= b 
  // ||ax - b|| / ||a|| <= r2  =>  ax + ||a||r2 >= b
  // 0.1 <= r1 <= r2
  // index starts from 1, there are 2*consNum constraints
  for(int i = 0; i < consNum; ++ i) {
    double constant = poly.GetConstant(i) ;
    glp_set_row_bnds(glpkinter._glp, 2*i+1, GLP_UP, 0.0, constant) ;
    glp_set_row_bnds(glpkinter._glp, 2*i+2, GLP_LO, constant, 0.0) ;  
  } 
  // there are two new column for r1 and r2
  glp_add_cols(glpkinter._glp, variNum+2) ;
  for(int j = 1; j < variNum+1; ++ j) {
    glp_set_col_bnds(glpkinter._glp, j, GLP_FR, 0.0, 0.0) ;
  }
  glp_set_col_bnds(glpkinter._glp, variNum+1, GLP_LO, 0.1, 0.0) ;
  glp_set_col_bnds(glpkinter._glp, variNum+2, GLP_LO, 0.1, 0.0) ;
  // obj = r2 - r1
  glp_set_obj_coef(glpkinter._glp, variNum+1, -1.0) ;
  glp_set_obj_coef(glpkinter._glp, variNum+2, 1.0) ;
  int valNum = 2 * consNum * (variNum + 2) ;
  int* idxi = new int[valNum+1] ;
  int* idxj = new int[valNum+1] ;
  double* vals = new double[valNum+1] ;
  int idx = 1 ;
  for(int i = 0; i < consNum; ++ i) {
    for(int j = 0; j < variNum; ++ j) {
      idxi[idx] = 2 * i + 1 ;
      idxj[idx] = j + 1 ;
      vals[idx] = poly.GetCoef(i, j) ;
      ++ idx ;
      idxi[idx] = 2 * i + 2 ;
      idxj[idx] = j + 1 ;
      vals[idx] = poly.GetCoef(i, j) ;
      ++ idx ;
    }
  } 
  for(int i = 0; i < consNum; ++ i) {
    double currNorm = poly.GetConstraint(i).norm() ;
    idxi[idx] = 2 * i + 1 ;
    idxj[idx] = variNum + 1 ;
    vals[idx] = currNorm ;
    ++ idx ;
    idxi[idx] = 2 * i + 1 ;
    idxj[idx] = variNum + 2 ;
    vals[idx] = 0 ;
    ++ idx ;
    idxi[idx] = 2 * i + 2 ;
    idxj[idx] = variNum + 1 ;
    vals[idx] = 0 ;
    ++ idx ;
    idxi[idx] = 2 * i + 2 ;
    idxj[idx] = variNum + 2 ;
    vals[idx] = currNorm ;
    ++ idx ;
  }
  glp_load_matrix(glpkinter._glp, valNum, idxi, idxj, vals) ;
  glp_smcp para ;
  glp_init_smcp(&para) ;
  para.msg_lev = GLP_MSG_OFF ;
  glp_simplex(glpkinter._glp, &para) ;
  Point point ;
  if (glp_get_prim_stat(glpkinter._glp) == GLP_FEAS) {
    Vector coord(variNum) ;
    for(int j = 0; j < variNum; ++ j) {
      coord(j) = glp_get_col_prim(glpkinter._glp, j+1) ;
    } 
    point.set_coordinates(coord) ;
  }

  delete[] idxi ;
  delete[] idxj ;
  delete[] vals ;
  return point ;
}

/*******************************************************************************
 * Calls simplex method in GLPK and compute a satisfied point, i.e. this point
 * satisfy all the constraints of the polyhedron
 * @para poly the given polyhedron
 * @return the satisfied point  
*******************************************************************************/
Point GlpkInterface::GetSatPoint (const std::vector<int>& headIdx, const Polyhedron& poly) {
  int consNum = headIdx.size() ;
  int variNum = poly.get_variable_num() ;
  int oriConsNum = glp_get_num_rows(_glp) ;
  glp_add_rows(_glp, consNum) ; 
  double constant ;
  for(int i = 0; i < consNum; ++ i) {
    constant = poly.GetConstant( headIdx[i] ) ;
    // minus _epsilon for getting a point inside the polyhedron
    if (oriConsNum + i == 0) {
      // the first constraint is Ax >= b + epsilon, others are Ax <= b - epsilon
      glp_set_row_bnds(_glp, 1, GLP_LO, constant+_epsilon, 0.0) ;
    }
    else {
      glp_set_row_bnds(_glp, oriConsNum+i+1, GLP_UP, 0.0, constant-_epsilon) ;
    }
  }
  int* idxj = new int[variNum+1] ;
  double* vals = new double[variNum+1] ;
  if (oriConsNum == 0) {
    glp_set_obj_dir(_glp, GLP_MIN) ;
    glp_add_cols(_glp, variNum) ;
    for(int j = 1; j < variNum+1; ++ j) {
      glp_set_col_bnds(_glp, j, GLP_FR, 0.0, 0.0) ;
    }
    // set the object as 0 to avoid the optimization phase
    glp_set_obj_coef(_glp, 1, 0.0) ; 
  }
  
  for(int i = 0; i < consNum; ++ i) {
    int consIdx = headIdx[i] ;
    for(int j = 0; j < variNum; ++ j) {
      idxj[j+1] = j + 1 ;
      vals[j+1] = poly.GetCoef(consIdx, j) ;
    }
    glp_set_mat_row(_glp, oriConsNum+i+1, variNum, idxj, vals) ;
  } 
  glp_smcp para ;
  glp_init_smcp(&para) ;
  para.msg_lev = GLP_MSG_OFF ;
  glp_simplex(_glp, &para) ;
  Point point ;
  if (glp_get_prim_stat(_glp) == GLP_FEAS) {
    Vector coord(variNum) ;
    for(int j = 0; j < variNum; ++ j) {
      coord(j) = glp_get_col_prim(_glp, j+1) ;
    } 
    point.set_coordinates(coord) ;
  }
 
  delete[] idxj ;
  delete[] vals ;
  return point ;
}


/*******************************************************************************
 * Tests if the {idx}th constraint is redundant.
*******************************************************************************/
bool GlpkInterface::Sat (const Polyhedron& poly, int idx) {
  GlpkInterface glpkinter ;
  int consNum = poly.get_constraints_num() ;
  int variNum = poly.get_variable_num() ;
  glp_add_rows(glpkinter._glp, consNum) ; 
  // the first constraint is Ax >= b + epsilon, others are Ax <= b - epsilon
  double constant ;
  for(int i = 0; i < consNum; ++ i) {
    constant = poly.GetConstant(i) ;
    glp_set_row_bnds(glpkinter._glp, i+1, GLP_UP, 0.0, constant-glpkinter._epsilon) ;
  }
  constant = poly.GetConstant(idx) ;
  glp_set_row_bnds(glpkinter._glp, idx+1, GLP_LO, constant+glpkinter._epsilon, 0.0) ;
  int* idxj = new int[variNum+1] ;
  double* vals = new double[variNum+1] ;
  glp_set_obj_dir(glpkinter._glp, GLP_MIN) ;
  glp_add_cols(glpkinter._glp, variNum) ;
  for(int j = 1; j < variNum+1; ++ j) {
    glp_set_col_bnds(glpkinter._glp, j, GLP_FR, 0.0, 0.0) ;
  }
  // set the object as 0 to avoid the optimization phase
  glp_set_obj_coef(glpkinter._glp, 1, 0.0) ;  
  for(int i = 0; i < consNum; ++ i) {
    for(int j = 0; j < variNum; ++ j) {
      idxj[j+1] = j + 1 ;
      vals[j+1] = poly.GetCoef(i, j) ;
    }
    glp_set_mat_row(glpkinter._glp, i+1, variNum, idxj, vals) ;
  } 
  glp_smcp para ;
  glp_init_smcp(&para) ;
  para.msg_lev = GLP_MSG_OFF ;
  glp_simplex(glpkinter._glp, &para) ;
  bool res ;
  if (glp_get_prim_stat(glpkinter._glp) == GLP_FEAS) {
    res = true ;
  }
  else {
    res = false ;
  }
 
  delete[] idxj ;
  delete[] vals ;
  return res ;
}
