#include <stdio.h>
#include <stdbool.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>

typedef void operation(void*);

static inline int min(int a, int b) {
  return a<b ? a : b;
}

void time_op(const char *name, operation *f, void *data) {
  struct timespec start, end;
  double total;
  
  printf("%s: ", name);
  fflush(stdout);
  clock_gettime(CLOCK_MONOTONIC, &start);
  f(data);
  clock_gettime(CLOCK_MONOTONIC, &end);
  total = (end.tv_nsec - start.tv_nsec)*1E-9 + (end.tv_sec - start.tv_sec);
  printf("%lf\n", total);
  fflush(stdout);
}

typedef double real;

extern void dgemm_(const char *transA,
		   const char *transB,
		   const int *m,
		   const int *n,
		   const int *k,
		   const double *alpha,
		   const double *A,
		   const int *ldA,
		   const double *B,
		   const int *ldB,
		   const double *beta,
		   double *restrict C,
		   const int *ldC);

extern real dnrm2_(const int *n,
		   const double *X,
		   const int *incx);

void mat_mult_ijk(const int m,
	          const int n,
	          const int p,
	          real a[restrict m][p],
	          const real b[m][n],
	          const real c[n][p]) {
  for(int i=0; i<m; i++) {
    for(int k=0; k<p; k++) {
      a[i][k] = 0.0;
    }
  }
#ifdef _OPENMP
#pragma omp parallel for
#endif
  for(int i=0; i<m; i++) {
    for(int j=0; j<n; j++) {
      for(int k=0; k<p; k++) {
	a[i][k] += b[i][j] * c[j][k];
      }
    }
  }
}

#define BLOCKING 64
void mat_mult_ijk_blocked(
		  const int m,
	          const int n,
	          const int p,
	          real a[restrict m][p],
	          const real b[m][n],
	          const real c[n][p]) {
  for(int i=0; i<m; i++) {
    for(int k=0; k<p; k++) {
      a[i][k] = 0.0;
    }
  }
#ifdef _OPENMP
  #pragma omp parallel for
#endif
  for(int block_i=0; block_i<m; block_i += BLOCKING) {
    for(int block_j=0; block_j<n; block_j += BLOCKING) {
      for(int i=block_i; i<min(block_i + BLOCKING, m); i++) {
	for(int j=block_j; j<min(block_j + BLOCKING, n); j++) {
	  real w = b[i][j];
	  for(int k=0; k<p; k++) {
	      a[i][k] += w * c[j][k];
	  }
	}
      }
    }
  }
}

void mat_mult_ijk_blocked_tasks(
		  const int m,
	          const int n,
	          const int p,
	          real a[restrict m][p],
	          const real b[m][n],
	          const real c[n][p]) {
  for(int i=0; i<m; i++) {
    for(int k=0; k<p; k++) {
      a[i][k] = 0.0;
    }
  }
#ifdef _OPENMP
  #pragma omp parallel
#endif
  {
    #pragma omp master
    {
      for(int block_i=0; block_i<m; block_i += BLOCKING) {
#ifdef _OPENMP
	#pragma omp task
#endif
	{
	  for(int block_j=0; block_j<n; block_j += BLOCKING) {
	    for(int i=block_i; i<min(block_i + BLOCKING, m); i++) {
	      for(int j=block_j; j<min(block_j + BLOCKING, n); j++) {
		real w = b[i][j];
		for(int k=0; k<p; k++) {
		  a[i][k] += w * c[j][k];
		}
	      }
	    }
	  }
	}
      }
    }
    #pragma omp taskwait
  }
}

void mat_mult_ikj(const int m,
	          const int n,
	          const int p,
	          real a[restrict m][p],
	          const real b[m][n],
	          const real c[n][p]) {
  for(int i=0; i<m; i++) {
    for(int k=0; k<p; k++) {
      a[i][k] = 0.0;
    }
  }
  for(int i=0; i<m; i++) {
    for(int k=0; k<p; k++) {
      for(int j=0; j<n; j++) {
	a[i][k] += b[i][j] * c[j][k];
      }
    }
  }
}

void mat_mult_ikj_transposed(const int m,
	          const int n,
	          const int p,
	          real a[restrict m][p],
	          const real b[m][n],
	          const real c[n][p]) {
  real ct[p][n];
  for(int i=0; i<m; i++) {
    for(int k=0; k<p; k++) {
      a[i][k] = 0.0;
    }
  }
  for(int j=0; j<n; j++) {
    for(int k=0; k<p; k++) {
      ct[k][j] = c[j][k];
    }
  }
  
  for(int i=0; i<m; i++) {
    for(int k=0; k<p; k++) {
      for(int j=0; j<n; j++) {
	a[i][k] += b[i][j] * ct[k][j];
      }
    }
  }
}

/* slow unless -ffast-math */
void mat_mult_ikj_transposed_blocked(const int m,
	          const int n,
	          const int p,
	          real a[restrict m][p],
	          const real b[m][n],
	          const real c[n][p]) {
  real ct[p][n];
  for(int j=0; j<n; j++) {
    for(int k=0; k<p; k++) {
      ct[k][j] = c[j][k];
    }
  }
  
  for(int block_i=0; block_i<m; block_i += BLOCKING) {
    for(int block_k=0; block_k<p; block_k += BLOCKING) {
      for(int i=block_i; i<min(block_i + BLOCKING, m); i++) {
	for(int k=block_k; k<min(block_k + BLOCKING, p); k++) {
	  double r = 0.0;
	  for(int j=0; j<n; j++) {
	    r += b[i][j] * ct[k][j];
	  }
	  a[i][k] = r;
	}
      }
    }
  }
}

void mat_mult_blas(const int cm,
	           const int cn,
	           const int cp,
	           real ca[restrict cm][cp],
	           const real cb[cm][cn],
	           const real cc[cn][cp]) {
  const int m = cp, n = cm, k = cn;
  const real alpha = 1.0, beta = 0.0;
  const real *a = (const real*) &cc[0][0];
  const real *b = (const real*) &cb[0][0];
  real *c = (real*) &ca[0][0];
  dgemm_("N", "N", &m, &n, &k, &alpha, a, &m, b, &k, &beta, c, &m);
}

bool equal_mat(const int m,
	       const int n,
	       const real a[m][n],
	       const real b[m][n]) {
  for(int i=0; i<m; i++) {
    for(int j=0; j<n; j++) {
      if (a[i][j] != b[i][j]) return false;
    }
  }
  return true;
}

double norm2_seq(const int m,
	         const real a[m]) {
  real x = 0.0;
  for(int i=0; i<m; i++) {
    real y = a[i];
    x += y*y;
  }
  return sqrt(x);
}

double norm2_par(const int m,
	         const real a[m]) {
  real x = 0.0;
#ifdef _OPENMP
#pragma omp parallel for reduction(+ : x)
#endif
  for(int i=0; i<m; i++) {
    real y = a[i];
    x += y*y;
  }
  return sqrt(x);
}

double norm2_blas(const int m,
	          const real a[m]) {
  const int incx = 1;
  return dnrm2_(&m, a, &incx);
}

void init_random(const int basis,
		 const int multiplier,
		 const int m,
		 const int n,
	         real a[m][n]) {
  int x = (int8_t) basis;
  for(int i=0; i<m; i++) {
    for(int j=0; j<n; j++) {
      a[i][j] = x;
      x = (int8_t) (x * multiplier);
    }
  }
}

#define M 1800
#define N 2200
#define P 1400

static real B[M][N], C[N][P];

#define OP_MAT_MULT(suffix) \
static real A_##suffix[M][P]; \
void op_mat_mult_##suffix(void* dummy) { \
  mat_mult_##suffix(M, N, P, A_##suffix, B, C); \
}

#define OP_NORM2(suffix) \
static real s_##suffix; \
void op_norm2_##suffix(void *dummy) { \
  s_##suffix = norm2_##suffix(M*N, &B[0][0]); \
}   

OP_MAT_MULT(ijk)
OP_MAT_MULT(ijk_blocked)
OP_MAT_MULT(ijk_blocked_tasks)
OP_MAT_MULT(ikj)
OP_MAT_MULT(ikj_transposed)
OP_MAT_MULT(ikj_transposed_blocked)
OP_MAT_MULT(blas)

OP_NORM2(seq)
OP_NORM2(par)
OP_NORM2(blas)

int main() {
  init_random(0x42, 0x43, M, N, B);
  init_random(0x88, 0x15, N, P, C);
  printf("matmult\n");
  time_op("ijk", op_mat_mult_ijk, NULL);
  time_op("ijk_blocked", op_mat_mult_ijk_blocked, NULL);
  printf("ijk = ijk_blocked: %d\n", equal_mat(M, P, A_ijk, A_ijk_blocked));
  time_op("ijk_blocked_tasks", op_mat_mult_ijk_blocked_tasks, NULL);
  printf("ijk = ijk_blocked_tasks: %d\n", equal_mat(M, P, A_ijk, A_ijk_blocked_tasks));
  /*time_op("ikj", op_mat_mult_ikj, NULL);
    printf("ijk = ikj: %d\n", equal_mat(M, P, A_ijk, A_ikj));
  time_op("ikj_transposed", op_mat_mult_ikj_transposed, NULL);
  printf("ijk = ikj_transposed: %d\n", equal_mat(M, P, A_ijk, A_ikj_transposed));
  time_op("ikj_transposed_blocked", op_mat_mult_ikj_transposed_blocked, NULL);
  printf("ijk = ikj_transposed_blocked: %d\n", equal_mat(M, P, A_ijk, A_ikj_transposed_blocked));*/
  time_op("blas", op_mat_mult_blas, NULL);
  mat_mult_blas(M, N, P, A_blas, B, C);
  printf("ijk = blas: %d\n", equal_mat(M, P, A_ijk, A_blas));
  printf("norm2\n");
  time_op("seq", op_norm2_seq, NULL);
  time_op("par", op_norm2_par, NULL);
  time_op("blas1", op_norm2_blas, NULL);
  printf("seq = blas1: %d\n", s_seq == s_blas);
  printf("seq = par: %d\n", s_seq == s_par);
  return 0;
}
